/*
 * MIT License
 *
 * Copyright (c) 2024 Todd Ginsberg
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package com.ginsberg.junit.exit.agent;

import com.ginsberg.junit.exit.ExitPreventerStrategy;
import com.ginsberg.junit.exit.SystemExitPreventedException;

@DoNotRewriteExitCalls
public class AgentSystemExitHandlerStrategy implements ExitPreventerStrategy {

    private static Integer firstExitStatusCode;
    private static boolean loadedFromAgent = false;
    private static boolean isRunningTest = false;

    public static void handleExit(final int status) {
        if(isRunningTest) {
            if (firstExitStatusCode == null) {
                firstExitStatusCode = status;
            }
            throw new SystemExitPreventedException(firstExitStatusCode);
        } else {
            System.exit(status);
        }
    }

    public static void agentInit() {
        loadedFromAgent = true;
    }

    public static boolean isLoadedFromAgent() {
        return loadedFromAgent;
    }

    @Override
    public Integer firstExitStatusCode() {
        return firstExitStatusCode;
    }

    @Override
    public void beforeTest() {
        isRunningTest = true;
    }

    @Override
    public void afterTest() {
        isRunningTest = false;
    }

    @Override
    public void resetBetweenTests() {
        firstExitStatusCode = null;
    }
}
